/* RWhois Software 

Copyright (c) 1994 Scott Williamson and Mark Kosters

Users may distribute the binary and source code to third parties provided
that the copyright notice and this statement appears on all copies and that
no charge is made for such copies.  

This software is provided AS IS with no warranties of any kind.  The authors
shall have no liability with respect to the infringement of copyrights,
trade secrets or any patents by this file or any part thereof.  In no
event will the authors be liable for any lost revenue or profits or
other special, indirect and consequential damages.

Comments and additions should be sent to the authors:

                    Scott Williamson
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4820
					scottw@internic.net

                    Mark Kosters
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4795
					markk@internic.net

File: pullup.c  Everything starts here. 
Developed by: Scott Williamson

*/


#define MAIN 1

#include "rwhois.h"
#include "error.h"

char *index();


main(argc, argv)
int argc;
char *argv[];
{
	int finished = FALSE;
	char **quicklook;
	struct SERVER *this_server;
	int wret = 0;  /* return from wait routine */
	char cmdline[MAXBUFFER];
	int cret = 0;	/* return from contact attempt */
	char rcfile[MAXBUFFER];	/* system wide default rc file */
	char lrcfile[MAXBUFFER];	/* users rc file */
	char *home_env;		/* users home directory..from the environment*/
	char *getenv();



	/* save the command name (usually called "rwhois") */
	cmdname = argv[0];

	argc--, argv++;

	printf("\nRWhois Pullup client --  RWhois spec version %s (InterNIC V-%s)\n",
		SPEC_VERSION, IMP_VERSION);

	if( argc < 1 )
	{
		help();
		exit(-1);
	} 

	shots = DEFAULTSHOT;  /* go through the entire refeal list */
	verbose = WHISPER;	
	default_wait = DEF_WAITTIME;
	default_maxwait = DEF_MAXWAIT;
	default_port = DEF_PORT;
	default_load = 0;
	send_limit( DEF_LIMIT, LIMIT_SET); /* set init default limit request */

	extreferral = FALSE;
	extsee_also = FALSE;
	default_cache = OFF;
	control_pannel[CP_CACHE] = SET_ALWAYS|SET_SEND;
	/* default_holdconnect = ON; */
	/* control_pannel[CP_HOLDCONNECT] = SET_ALWAYS|SET_SEND; */
	forcing = FALSE;
	authority = FALSE;
	current_modes = MODE_NONE;
	to_file = TRUE;
	f_out = NULL;
	goodedit = REG_NONE;
	edit = OFF;
	display_cap = DISPLAY_NONE;
	punt_referral = FALSE;
	user = NULL;
	authstring = NULL;
	just_show_config = FALSE;
	preserve_temp = FALSE; /* diag flag to preserve temp files -+ option*/

	/* setup the sever base structure */
	SERVERINIT;



	/* look quickly thourgh the augv to see if things should be verbose */
	/* don't act on all options till defaults are setup                 */

	quicklook = argv;

	while( *quicklook)
	{
		if (!strcmp(*quicklook, "-v")) 
			verbose = TALK;
		else if (!strcmp(*quicklook, "-d"))
			verbose = SHOUT;
		else if(!strcmp(*quicklook, "-q"))
			verbose = QUIET; 
		++quicklook;
	}


	/* seed the root server as default servers */
	/* in final this should not be */

 	cmdline_server( SERDEF1, REF_PRESET);



	/**************************************************************/
	/* grab the defaults frist then process the command line  ..  */
	/* command line overrides defaults ... you will get a chance  */
	/* to run a different config file to modify defaults again    */
	/**************************************************************/


	/* first try the system wide rc file */
	sprintf( rcfile, "%s/rwhois.rc", CONFPATH);
	/* Try to process it */
	procrc( rcfile);

	/* Now try to open the user rc file and override anything needed */
	/* use the HOME environmential variable */
	if(!(home_env = getenv( "HOME")))
	{
		if( verbose > TALK )
			error("No HOME environment variable");
	}
	else
	{
		sprintf( lrcfile, "%s/.rwhois.rc", home_env);
		procrc( lrcfile);
	}

	/* select options and build the command list for whois/rwhois server */
	procargs( argv, cmdline);

	/* update the query for each of the default servers */
	upd_query( cmdline);

	if( !forcing )
		control_pannel[CP_QUERY] = (SET_SEND|SET_ALWAYS);

	/* if verbose mode show the user the final configuration */
	if( verbose > WHISPER || just_show_config)
	{
		printf("\nFinal configuration list:");
		list_defaults();
	}

	/* Turn off any contol display send */
	control_pannel[CP_DISPLAY] = 0; 

	if( just_show_config)
		exit( 0);

	current_referral = 0;


	/* set server pointer to current server */
	this_server = SERVERCURRGET;

	/*  Start of the main query loop */

	while( !finished  )
	{
		if((cret = contact( this_server)) == CONTGOOD)
		{

			/* server after connection */
			this_server = SERVERCURRGET;

			if( verbose > WHISPER)
			{
				printf("\nConnnection made to %s port %d connection took %s\n",
					this_server->server, this_server->port, 
					showtime( this_server->start_time, this_server->con_time));
			}


			if( wait_result( default_wait) == WAIT_IDYES)
			{
				server_setup();
				send_xfer( "referral");
				wret = wait_result( default_wait);


				closeout( this_server);

				procreferrals( this_server);
			}

		}

		if( !(this_server = SERVERCURRNEXT))
			finished = TRUE;

	}

	server_list( LISTALL, stdout);

	fflush(stdout);
	return 0;
}



procreferrals( proc_server)
struct SERVER *proc_server;
{
FILE *fp;
char buffer[MAXBUFF];
char temp[MAXBUFF];
char *ptr;


	if(!(proc_server->filestatus&FILEHASDATA))
		return(ERROR);


	if(!(fp = fopen( proc_server->filename, "r")))
	{
		sprintf( temp, "Unable to open display file [%s]", 
			proc_server->filename);
		error( temp);
	}
	else
	{
		if( verbose > TALK )
		{
			sprintf( buffer, "Processing results from [%s]", 
				proc_server->filename);
			message( buffer);
		}

		while(fgets( buffer, 256, fp))
		{
			if(strncmp( buffer, "#RWhois:", 7))
			{
				if(!strncmp( buffer, "referral", 8))
				{
					ptr = index( buffer, ':');
					if(!strncmp( ++ptr, "Referral", 8))
					{
						ptr = index( ptr, ':');
						add_server( ++ptr);
					}
				}
			}
		}
		fclose( fp);
	}

	if( !preserve_temp)
  		unlink( proc_server->filename);

	return( GOOD);
}


add_server( sstring)
char *sstring;
{
char *host;
int port, type;
struct SERVER bld_server;
char *refhost;
struct SERVER *curr_server;
char temp[MAXBUFF];

	curr_server = SERVERCURRGET;

	refhost = safecalloc( 1, strlen( curr_server->server) + 5);
	sprintf( refhost, "%s:%d", curr_server->server, curr_server->port);
			

			brk_server( sstring, &port, &type);

			sprintf( temp, "Adding [%s] port [%d] type [%d]",
				 sstring, port, type);
			note( temp);

			bld_server.port = port;
			bld_server.type_status = type;
			bld_server.server = safecalloc( 1, strlen( sstring));
			strcpy( bld_server.server, sstring);
			bld_server.query = safecalloc( 1,1);
			bld_server.reftype = REF_REFERRAL;
			bld_server.referral = refhost;

			if(!server_main( SER_COMPARE, &bld_server))
				server_main( SER_ADD, &bld_server);
			else 
				printf("\nReceived repeat referral not adding [%s]", sstring);

			free( bld_server.server);
			free( refhost);



}
