/*
Copyright (c) 1994 Scott Williamson and Mark Kosters

Users may distribute the binary and source code to third parties provided
that the copyright notice and this statement appears on all copies and that
no charge is made for such copies.  

This software is provided AS IS with no warranties of any kind.  The authors
shall have no liability with respect to the infringement of copyrights,
trade secrets or any patents by this file or any part thereof.  In no
event will the authors be liable for any lost revenue or profits or
other special, indirect and consequential damages.

Comments and additions should be sent to the authors:

                    Scott Williamson
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4820
                    scottw@internic.net

                    Mark Kosters
                    Network Solutions
                    505 Huntmar Park Drive
                    Herndon, VA 22071
                    (703) 742-4795
                    markk@internic.net

File: server.c     Contains all of the RWhois server support functions.
Developed by: Scott Williamson

*/

#include "rwhois.h"
#include "error.h"


static struct SERVER *curr_server, *ret_server, *head_server, *mark_server;
static struct SERVER *mod_server;
static struct SERVER *intmark_server, *holdnext_server, *holdprev_server;
static struct SERVER *ins_server;
static int count;

struct SERVER *server_main( action, server_in)
  int action;
struct SERVER *server_in;
{
  char *hold_name, *hold_query;
  long marktime();


  ret_server = NULL;

  switch( action)
  {

  case SER_REWIND:

    curr_server = head_server;
    while( curr_server)
    {
      free( curr_server->server);
      free( curr_server->query);
      holdnext_server = curr_server->next_server;
      free( curr_server);
      curr_server = holdnext_server;
    }

    /* must call init or things won't work right */

  case SER_INIT:

    head_server = (struct SERVER *) safecalloc( 1, sizeof( struct SERVER));
    ret_server = ins_server = intmark_server = mark_server 
               = curr_server = head_server;
    head_server->next_server = NULL;
    head_server->prev_server = NULL;
    count = referral = 0;
    break;

  case SER_ERASE:

    holdnext_server = curr_server->next_server;
    curr_server->next_server = NULL;
    ins_server = intmark_server = mark_server = curr_server;

    while( holdnext_server)
    {
      free( holdnext_server->query);
      holdprev_server = holdnext_server->next_server;
      free( holdnext_server);
      --count; --referral;
      holdnext_server = holdprev_server;
    }

    break;

                

  case SER_COMPARE:

    holdnext_server = head_server;
    ret_server = NULL;

    while( holdnext_server && !ret_server)
    {
      if((!strcmp( holdnext_server->server, server_in->server)
          && holdnext_server->port == server_in->port))
      {
        if( holdnext_server->query)
        {
          if(!strncmp(holdnext_server->query, 
                      server_in->query, strlen( server_in->query)))
          {
            ret_server = holdnext_server;
          }
        }   
        else
          ret_server = holdnext_server;
      }
      holdnext_server = holdnext_server->next_server;
    }
    break;

  case SER_ADD:

    if( count)
    {
      intmark_server->next_server =
                                   (struct SERVER * ) safecalloc( 1, sizeof( struct SERVER));
      intmark_server->next_server->prev_server = intmark_server;
      intmark_server = intmark_server->next_server;
    }
    intmark_server->server = safecalloc( 1,(strlen( server_in->server)+1));
    strcpy( intmark_server->server, server_in->server);
    intmark_server->port = server_in->port;
    intmark_server->type_status = server_in->type_status;
    intmark_server->reftype = server_in->reftype;

    if( server_in->query)
    {
      intmark_server->query = safecalloc( 1, (strlen( server_in->query)+1));
      strcpy( intmark_server->query, server_in->query);
    }
    else
      intmark_server->query = NULL;

    intmark_server->reftype = server_in->reftype;
    if( server_in->referral)
    {
      intmark_server->referral = safecalloc( 1, strlen(server_in->referral)+1);
      strcpy( intmark_server->referral , server_in->referral);
    }
    else
      intmark_server->referral = NULL;
    intmark_server->server_id = count+1;
    ++count; ++referral;
    break;

  case SER_MODIFY:


    /* hold server name in case pointer comes from here */


    mod_server = head_server;

    while( mod_server && (mod_server->server_id 
                          != server_in->server_id))
      mod_server = mod_server->next_server;

    if( mod_server)
    {

      hold_name = safecalloc( 1, strlen( server_in->server)+1);
      strcpy( hold_name, server_in->server);
      free( mod_server->server); 


      /* hold query incase pointer comes from here */
      if( server_in->query)
      {
        hold_query = safecalloc( 1, strlen( server_in->query)+1);
        strcpy( hold_query, server_in->query);
        free( mod_server->query);

      }

      bcopy( mod_server, server_in, sizeof( struct SERVER));

      mod_server->server = (char *)safecalloc( 1,
                                               (strlen( hold_name)+1));
      strcpy( mod_server->server , hold_name);
      free( hold_name);

      mod_server->query = safecalloc( 1, 
                                      (strlen( hold_query)+1));
      strcpy( mod_server->query, hold_query);
      free( hold_query);

    }
    else
      printf("\nModify for server not found");

    break;

  case SER_DELETE:

    holdprev_server = curr_server->prev_server;
    holdnext_server = curr_server->next_server;
    if( holdnext_server)
      holdnext_server->prev_server = curr_server->prev_server;

    if( holdprev_server)
      holdprev_server->next_server = holdnext_server;

    free( curr_server->server);
    free( curr_server);

    ins_server = curr_server = holdnext_server;
    if( count)
    {
      --count;
      --referral;
    }
    break;

  case SER_INSERT:
    holdprev_server = ins_server;
    holdnext_server = ins_server->next_server;

    if( count)
    {
      ins_server->next_server =
                               (struct SERVER * ) safecalloc( 1, sizeof( struct SERVER));
      ins_server = ins_server->next_server;
    }

    ins_server->server = safecalloc( 1,(strlen( server_in->server)+1));
    strcpy( ins_server->server, server_in->server);
    ins_server->port = server_in->port;
    ins_server->type_status = server_in->type_status;
    ins_server->reftype = server_in->reftype;

    if( server_in->query)
    {
      ins_server->query = safecalloc( 1, (strlen( server_in->query)+1));
      strcpy( ins_server->query, server_in->query);
    }
    else
      ins_server->query = NULL;

    ins_server->next_server = holdnext_server;
    ins_server->prev_server = holdprev_server;
    ins_server->reftype = server_in->reftype;

    if( server_in->referral)
    {
      ins_server->referral = safecalloc( 1, strlen(server_in->referral));
      strcpy( ins_server->referral , server_in->referral);
    }
    else
      ins_server->referral = NULL;

    intmark_server->server_id = count+1;
    ++count; ++referral;
    break;

  case SER_CURRGET:
    ret_server = curr_server;
    break;

  case SER_CURRNEXT:
    if( curr_server->next_server)
    {
      ins_server = ret_server = curr_server = curr_server->next_server;
      ++current_referral;
    }
    else
      ret_server = NULL;
    break;

  case SER_CURRPREV:
    if( curr_server->prev_server)
    {
      ins_server = ret_server = curr_server = curr_server->prev_server;
      --current_referral;
    }
    else
      ret_server = NULL;

  case SER_CURRHEAD:
    ret_server = ins_server = curr_server = head_server;
    current_referral = 0;
    break;

  case SER_MARK:
    ret_server = mark_server = curr_server;
    break;

  case SER_MARKNEXT:

    if( mark_server->next_server )
      ret_server = mark_server = mark_server->next_server;
    else
      ret_server = NULL;
    break;

  case SER_MARKHEAD:
    ret_server = mark_server = head_server;
    break;

  case SER_MARKPREV:

    if( mark_server->prev_server )
      ret_server = mark_server = mark_server->prev_server;
    else
      ret_server = NULL;
    break;

  case SER_MARKGET:
    ret_server = mark_server;
    break;

  case SER_STARTTIME:
    curr_server->start_time = marktime();
    break;

  case SER_CONTIME:
    curr_server->con_time = marktime();
    break;
        
  case SER_ANSTIME:
    curr_server->ans_time = marktime();
    break;

  default:
    printf("\nBad server maintenance option\n");
    ret_server = NULL;
    break;
  }

#ifdef DEBUG

  holdnext_server = head_server;
  printf("\n\nACTION = ");
  switch( action)
  {
  case( SER_ADD):
    printf("ADD");
    break;
  case( SER_MODIFY):
    printf("Modify");
    break;
  case( SER_REWIND):
    printf("REWIND");
    break;
  case( SER_ERASE):
    printf("ERASE");
    break;
  case( SER_COMPARE):
    printf("COMPARE");
    break;
  }

  if( count && ( action == SER_ADD || action == SER_MODIFY 
                 || action == SER_INSERT ))
  {
    printf(" Count = %d",count);    

    while( holdnext_server)
    {
      if( holdnext_server == curr_server )
        printf("\n**CURRENT**");
      if( holdnext_server == ins_server)
        printf("\n**INSERT**");
      if( holdnext_server == mark_server)
        printf("\n**MARK**");
      if( holdnext_server == intmark_server)
        printf("\n**INT**");
      printf("\nID = %d server = [%s] port = %d  type = %d", 
             holdnext_server->server_id, holdnext_server->server,
             holdnext_server->port, holdnext_server->type_status);
      fflush(stdout);
      printf("\nstart = %ld con = %ld ans = %ld", holdnext_server->start_time,
             holdnext_server->con_time, holdnext_server->ans_time);
      fflush(stdout);
      printf("\nquery = [%s] referral type = %d", 
             holdnext_server->query ? holdnext_server->query : "none",  
             holdnext_server->reftype);
      fflush(stdout);
      holdnext_server = holdnext_server->next_server;
    }
  }

#endif

  return( ret_server);
}


/***************************
struct SERVER {
    int server_id;
    char *server;
    int sh;
    int port;
    int reftype;
    int type_status; 
    char *referral;
    char *query;
    char *filename;
    char *bkupname;
    int filestatus;
    FILE *fp;
    FILE *bkfp;
    time_t  start_time; 
    time_t  con_time;
    time_t  ans_time
    struct STATUS server_status;
    struct SERVER *next_server;
    struct SERVER *prev_server;
};

***********************/

server_list(listmode, fp)
  int listmode;
FILE *fp;
{
  struct SERVER *list_server;

  if( listmode == LISTALL )
    SERVERMARKHEAD;
  else
    SERVERMARK;

  fprintf( fp, "\nCurrent RWhois server list:[%d:%d]:\n",
           referral, current_referral);

  if( list_server = SERVERMARKGET)
  {
    do{
      fprintf( fp, "[%d] %s port %d -> ", list_server->server_id,
               list_server->server, list_server->port);


      switch((list_server->reftype&0xff))
      {
      case REF_CONFFILE:
        fprintf( fp, "Config file");
        break;
      case REF_REFERRAL:
        fprintf( fp, "Referral");
        break;
      case REF_SEEALSO:
        fprintf( fp, "See-also");
        break;
      case REF_PRESET:
        fprintf( fp, "Preset");
        break;
      case REF_CMDLINE:
        fprintf( fp, "Command Line");
        break;
      default:
        fprintf( fp, "Unknown");

      }

      fprintf( fp, " From [%s] ", list_server->referral ? list_server->referral : "Unknown");
                
      fprintf( fp, " [%s]... Type: ", list_server->query ? list_server->query :
               "Not specified" );

      switch( list_server->type_status & 0x0f )
      {
      case TYPE_RWHOIS:
        fprintf( fp, "RWhois\n");
        break;
      case TYPE_WHOIS:
        fprintf( fp, "Whois\n");
        break;
      case TYPE_WHOISPLUSPLUS:
        fprintf( fp, "Whois++\n");
        break;
      default:
        fprintf( fp, "Unknown\n");
        break;
      }

      if( listmode == LISTBACK)
        list_server = SERVERMARKPREV;
      else
        list_server = SERVERMARKNEXT;

    }while(list_server);
  }
}


brk_server( string_in, port_out, type_out)
  char *string_in;
int *port_out, *type_out;
{
  char *ptr_new_port, *ptr_new_type;

  if( ptr_new_port = index( string_in, ':'))
  {
    *ptr_new_port = NullCH;
    ++ptr_new_port;
    if( ptr_new_type = index( ptr_new_port, ':'))
    {
      *(ptr_new_type++) = NullCH;
      *type_out = inter_type( ptr_new_type);
    }
    else
      *type_out = TYPE_UNKNOWN;

    *port_out = atoi( ptr_new_port);
  }
  else
  {
    *port_out = NORMPORT; 
    *type_out = TYPE_UNKNOWN;
  }

}

cmdline_server( sstring, reftype)
  char *sstring;
int reftype;
{
  int port, type;
  struct SERVER *tmp_server;
  char *buffer;

  tmp_server = ( struct SERVER * ) safecalloc( 1, sizeof( struct SERVER));

  /* working buffer copy */
  buffer = safecalloc( 1, (strlen( sstring)+1));
  strcpy( buffer, sstring);

  /* break-up the host:port:type string */
  brk_server( buffer, &port, &type);

  /* allocate and fill working server structure */
  tmp_server->server = safecalloc( 1, strlen( buffer)+1);

  strcpy( tmp_server->server ,buffer);
  tmp_server->port = port;
  tmp_server->type_status = type;
  tmp_server->query = NULL;
  tmp_server->reftype = reftype;

  tmp_server->referral = NULL;

  /* add the server as next in list*/
  server_main( SER_ADD, tmp_server);

  free( tmp_server->server);
  free( tmp_server);

}

inter_type( type)
  char *type;
{
  int ret;

  upper( type);

  if(!strcmp( type, "WHOIS"))
    ret = TYPE_WHOIS;
  else if(!strcmp( type, "RWHOIS"))
    ret = TYPE_RWHOIS;
  else if(!strcmp( type, "WHOIS++"))
    ret = TYPE_WHOISPLUSPLUS;
  else if(!strcmp( type, "OTHER"))
    ret = TYPE_OTHER;
  else
    ret = TYPE_UNKNOWN;

  return( ret);

}

upd_query( query)
  char *query;
{
  struct SERVER *this_server;



  if((this_server = SERVERMARKHEAD))
  {

    do{
      this_server->query = safecalloc( 1, strlen( query)+1);
      strcpy( this_server->query, query);
    }while((this_server = SERVERMARKNEXT));
  }
}

upd_status( server_num, stat_in)
  int server_num;
struct STATUS *stat_in;
{
  struct SERVER *this_server;
  struct SERVER *find_server();


  if((this_server = find_server( server_num)))
  {
    this_server->server_status.limit = stat_in->limit;
    this_server->server_status.load = stat_in->load;
    this_server->server_status.cache = stat_in->cache;
    this_server->server_status.holdconnect = stat_in->holdconnect;
    this_server->server_status.forward = stat_in->forward;
    this_server->server_status.authority = stat_in->authority;
    this_server->server_status.cached = stat_in->cached;
  }
  else
    return( -1);

  return(0);
}

prt_server_status()
{
  struct SERVER *stat_server;

  stat_server = SERVERCURRGET;

  printf("\nStatus of current server <%s> :", stat_server->server);
  printf("\nLimit       : %d", stat_server->server_status.limit);
  printf("\nLoad        : %f", stat_server->server_status.load);
  printf("\nCache       : %s", 
         stat_server->server_status.cache ? "OFF" : "ON");
  printf("\nHoldconnect : %s", 
         stat_server->server_status.holdconnect ? "OFF" : "ON");
  printf("\nForward     : %s", 
         stat_server->server_status.forward ? "OFF" : "ON");
  printf("\nAuthority   : %d", stat_server->server_status.authority);
  printf("\nCache       : %d\n", stat_server->server_status.cached);

}

struct SERVER *find_server( server_id)
  int server_id;
{
  struct SERVER *this_server;

  this_server = head_server;
  while( this_server && (this_server->server_id 
                         != server_id))
    this_server = this_server->next_server;

  return( this_server);
}

/* Finds a server that matches name,port,query from current server session */

struct SERVER *findlink( server, port, query)
  char *server;
int port;
char *query;
{
  struct SERVER *this_server;

  this_server = head_server;

  while( this_server )
  {
    if(!strcmp( this_server->server, server) 
       && !strcmp( this_server->query, query) 
       && this_server->port == port)
      return( this_server);

    this_server = this_server->next_server;
  }
  return( NULL);
}


server_setup()
{
  int x;
  int pass = 0;
  int ret = 0;
  struct SERVER *curr_server;

  /* process the setup directive the number order in rwhois.h  */
  /* is important to makesure that PRECMD is last */

  curr_server = SERVERCURRGET;

  for( x=0; x <  PANNEL_COUNT; ++x)
  {
    if( ((control_pannel[x]&SET_SEND) && !(control_pannel[x]&SET_FORCE))
        || (((control_pannel[x]&SET_SEND) 
             && (control_pannel[x]&SET_FORCE)) && pass))
    {

      do{
        ret = 0;

        switch( x)
        {

        case CP_HOLDCONNECT:
          send_command( "HOLDCONNECT ON");
          control_pannel[CP_HOLDCONNECT] =
                                          (control_pannel[CP_HOLDCONNECT]&~SET_SEND)|SET_WAIT;
          break;
        case CP_LIMIT:
          send_limit( 0, LIMIT_SEND);
          control_pannel[CP_LIMIT] = 
                                     (control_pannel[CP_LIMIT]&~SET_SEND)|SET_WAIT;
          break;
        case CP_CACHE:
          send_cache( default_cache);
          control_pannel[CP_CACHE] = 
                                     (control_pannel[CP_CACHE]&~SET_SEND)|SET_WAIT;
          break;
        case CP_SOA:
          if( autharea)
          {
            send_soa( autharea);
            control_pannel[CP_SOA] = 
                                     (control_pannel[CP_SOA]&~SET_SEND)|SET_WAIT;
          }
          else if( curr_server->query)
          {
            send_soa( curr_server->query);
            control_pannel[CP_SOA] = 
                                     (control_pannel[CP_SOA]&~SET_SEND)|SET_WAIT;

          }
          else
            control_pannel[CP_SOA] =
                                    (control_pannel[CP_SOA]&~SET_SEND);
          break;
        case CP_OBJECT:
          send_object( object_name);
          control_pannel[CP_OBJECT] = 
                                      (control_pannel[CP_OBJECT]&~SET_SEND)|SET_WAIT;
          break;
        case CP_XFER:
          send_xfer( xfer_object);
          control_pannel[CP_XFER] = 
                                    (control_pannel[CP_XFER]&~SET_SEND)|SET_WAIT;
          break;
        case CP_LOAD:
          send_command( "LOAD");
          control_pannel[CP_LOAD] =
                                   (control_pannel[CP_LOAD]&~SET_SEND)|SET_WAIT;
          break;
        case CP_SCHEMA:
          send_schema(  schema_object);
          control_pannel[CP_SCHEMA] =
                                     (control_pannel[CP_SCHEMA]&~SET_SEND)|SET_WAIT;
          break;
        case CP_DIRECTIVE:
          send_directive(  directive_name);
          control_pannel[CP_DIRECTIVE] =
                                        (control_pannel[CP_DIRECTIVE]&~SET_SEND)|SET_WAIT;
          break;
        case CP_STATUS:
          send_command(  "-STATUS");
          control_pannel[CP_STATUS] =
                                     (control_pannel[CP_STATUS]&~SET_SEND)|SET_WAIT;
          break;
        case CP_DEFINE:
          send_define( define_name );
          control_pannel[CP_DEFINE] =
                                     (control_pannel[CP_DEFINE]&~SET_SEND)|SET_WAIT;
          break;
        case CP_DISPLAY:

          /* must allow for multi dispaly sends */
          /* think we may need to shift to single config send */
          if(!(ret = send_display( NULL)))
          {
            control_pannel[CP_DISPLAY] =
                                        (control_pannel[CP_DISPLAY]&~SET_SEND)|SET_WAIT;
          }
          else
            control_pannel[CP_DISPLAY] =
                                        control_pannel[CP_DISPLAY]|SET_WAIT;
          break;
        case CP_PRECMD:
          send_extended( precmd);
          control_pannel[CP_PRECMD] =
                                     (control_pannel[CP_PRECMD]&~SET_SEND)|SET_WAIT;
          break;
        }

        if( control_pannel[x] & SET_WAIT)
          /* the while loop will allow limit neg  ** prob not best way ** */
          while( wait_result( default_wait) == WAIT_ERROR 
                 && err_num == EXCEEDED_MAXREC );

      } while( ret);

    }

    /* force a second pass for all forced directives */
    if( x == (PANNEL_COUNT-1) && !pass)
    {
      pass = TRUE;
      x = 0;
    }
  }
}
